/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package circlehunterserverlive;

import com.sun.javafx.functions.Function0;
import com.sun.javafx.functions.Function1;
import com.sun.javafx.functions.Function2;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import javafx.lang.FX;

/**
 *
 * @author thomas
 */
public class Server {

  private Function1<Void, Integer> clientConnected;
  private Function1<Void, Integer> clientDisconnected;

  private Function2<Void, Integer, String> messageReceived;
  private ServerSocket serverSocket;
  private int nextClientID = 1;
  private Map<Integer, ClientTask> clients = new ConcurrentHashMap<Integer, ClientTask>();

  public Server(Function1<Void, Integer> clientConnected,
          Function1<Void, Integer> clientDisconnected,
                Function2<Void, Integer, String> messageReceived) {


    this.clientConnected = clientConnected;
    this.clientDisconnected = clientDisconnected;
    this.messageReceived = messageReceived;
  }

  public void broadcast(String s) {
    for (ClientTask c : clients.values()) {
      send(c.id, s);
    }
  }

  public void start(int port) throws IOException {
    if (serverSocket != null) {
      throw new IllegalStateException("Server already started");
    }
    serverSocket = new ServerSocket(port);
    acceptClientsThread.start();
  }

  public boolean send(int id, String s) {
    OutputStream os = null;
    try {
      ClientTask t = clients.get(id);
      if (t == null) {
        throw new IllegalArgumentException("unknown client id " + id);
      }
      os = t.socket.getOutputStream();
      PrintWriter pw = new PrintWriter(os, true);
      pw.println(s);
      pw.flush();
    } catch (IOException ex) {
      Logger.getLogger(Server.class.getName()).log(Level.SEVERE, null, ex);
      return false;
    }
    return true;
  }

  private Thread acceptClientsThread = new Thread() {

    @Override
    public void run() {
      while (true) {
        try {
          Socket client = serverSocket.accept();
          final int id = nextClientID++;
          ClientTask t = new ClientTask(id, client);
          clients.put(id, t);
          FX.deferAction(new Function0<Void>() {
            @Override
            public Void invoke() {
              clientConnected.invoke(id);
              return null;
            }

          });
          t.start();

        } catch (IOException ex) {
          Logger.getLogger(Server.class.getName()).log(Level.SEVERE, null, ex);
        }
      }
    }
  };

  private class ClientTask extends Thread {
    public final int id;
    public final Socket socket;
    private static final long MIN_TIME_BETWEEN_MESSAGES = 50;

    public ClientTask(int id, Socket socket) {
      this.id = id;
      this.socket = socket;
    }

    @Override
    public void run() {
      try {
        BufferedReader r = new BufferedReader(new InputStreamReader(socket.getInputStream()));
        String s;
        while ((s = r.readLine()) != null) {
          final String message = s;
          FX.deferAction(new Function0<Void>() {
            @Override
            public Void invoke() {
              messageReceived.invoke(id, message);
              return null;
            }
          });
          try {
            Thread.sleep(MIN_TIME_BETWEEN_MESSAGES);
          } catch (InterruptedException ex) {
            Logger.getLogger(Server.class.getName()).log(Level.SEVERE, null, ex);
          }
        }

      } catch (IOException ex) {
        Logger.getLogger(Server.class.getName()).log(Level.SEVERE, null, ex);
      } finally {
        FX.deferAction(new Function0<Void>() {
            @Override
            public Void invoke() {
              clientDisconnected.invoke(id);
              return null;
            }
          });
      }
    }
  }
}
